#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "kp_core.h"
#include "kp_inference.h"
#include "helper_functions.h"

#include "kl720_demo_customize_inf_single_model.h"

static char _model_file_path[128] = "../../res/models/KL720/YoloV5s_640_640_3/models_720.nef";
//static char _model_file_path[128] = "../../res/models/KL720/yolov7/yolov7_tiny_person_head_TV_reparam_2112_448_id_286.nef";
//static char _model_file_path[128] = "../../res/models/KL720/yolov7/yolov7_tiny_person_head_tv_2112_448_id_211.nef";
//static char _image_file_path[128] = "../../res/images/one_bike_many_cars_608x608.bmp";
static char _image_file_path[128] = "../../res/images/people_talk_in_street_640x640.bmp";
//static char _image_file_path[128] = "./test_2_.bmp";
static int _loop = 1;

int main(int argc, char *argv[])
{
    kp_devices_list_t *device_list;
    kp_device_group_t device;
    kp_model_nef_descriptor_t model_desc;

    // each device has a unique port ID, 0 for auto-search
    int port_id = (argc > 1) ? atoi(argv[1]) : 0;
    int ret;

    /******* check the device USB speed *******/
    {
        int link_speed;
        device_list = kp_scan_devices();

        helper_get_device_usb_speed_by_port_id(device_list, port_id, &link_speed);
        if (KP_USB_SPEED_SUPER != link_speed)
            printf("[warning] device is not run at super speed.\n");
    }

    /******* connect the device *******/
    {
        int error_code;

        // connect device
        device = kp_connect_devices(1, &port_id, &error_code);
        if (!device)
        {
            printf("error ! connect device failed, port ID = '%d', error = '%d'\n", port_id, error_code);
            exit(0);
        }

        printf("connect device ... OK\n");

        kp_set_timeout(device, 5000);
    }

    /******* upload model to device *******/
    {
        ret = kp_load_model_from_file(device, _model_file_path, &model_desc);
        if (ret != KP_SUCCESS)
        {
            printf("error ! upload model failed, error = %d (%s)\n", ret, kp_error_string(ret));
            exit(0);
        }

        printf("upload model ... OK\n");
    }

    /******* prepare the image buffer read from file *******/
    // here convert a bmp file to RGB565 format buffer
    int img_width, img_height;
//    char *img_buf = helper_bmp_file_to_raw_buffer(_image_file_path, &img_width, &img_height, KP_IMAGE_FORMAT_RGB565);
    char *img_buf = helper_bmp_file_to_raw_buffer(_image_file_path, &img_width, &img_height, KP_IMAGE_FORMAT_RGBA8888);
    if (!img_buf)
    {
        printf("error ! read image file failed\n");
        exit(0);
    }

    int img_buf_size;
//    int img_width = 2112
//    int img_height = 448
    int img_channel = 4;

    float *pre_processing_buf;
    int32_t pre_processing_buf_size;

    int8_t *re_layout_buf;
    int32_t re_layout_buf_size;

    // read data file
//    img_buf = (char*)input_img.data;
    img_buf_size = img_width * img_height * img_channel;
//    printf("read image ... %s\n", (img_buf) ? "OK" : "failed");

    // get model input radix, scale
    kp_tensor_descriptor_t *input_node_0 = &(model_desc.models[0].input_nodes[0]);
    uint32_t model_input_radix = input_node_0->quantization_parameters.quantized_fixed_point_descriptor[0].radix;
    uint32_t model_input_scale = input_node_0->quantization_parameters.quantized_fixed_point_descriptor[0].scale;

    // get model input data layout
    uint32_t model_input_data_layout = input_node_0->data_layout;

    // get model input size (shape order: BxCxHxW)
    uint32_t model_input_channel = input_node_0->shape_npu[1];
    uint32_t model_input_height = input_node_0->shape_npu[2];
    uint32_t model_input_width = input_node_0->shape_npu[3];
//    printf("model_input_channel : %d\n",model_input_channel);
    // do normalization - this model is trained with normalize method: (data - 128) / 256
    pre_processing_buf_size = img_buf_size * sizeof(float);
    pre_processing_buf = (float *)malloc(pre_processing_buf_size);

    for (int i = 0; i < img_buf_size; i++) {
        pre_processing_buf[i] = ((float)img_buf[i] - 128.0) / 256.0;
//        pre_processing_buf[i] = ((float)img_buf[i]) / 255.0;
    }

//    free(img_buf);

    // toolchain calculate the radix value from input data (after normalization), and set it into NEF model.
    // NPU will divide input data "2^radix" automatically, so, we have to scaling the input data here due to this reason.
    float quantization_factor = powf(2, model_input_radix) * model_input_scale;

    for (int i = 0; i < img_buf_size; i++) {
        pre_processing_buf[i] *= quantization_factor;
    }

    // re-layout the data to fit NPU data layout format
    // KL720 supported NPU input layout format: 4W4C8B, 1W16C8B, 16W1C8B
    uint32_t width_align_base;
    uint32_t channel_align_base;

    if (KP_MODEL_TENSOR_DATA_LAYOUT_4W4C8B == model_input_data_layout) {
        width_align_base = 4;
        channel_align_base = 4;
    } else if (KP_MODEL_TENSOR_DATA_LAYOUT_1W16C8B == model_input_data_layout) {
        width_align_base = 1;
        channel_align_base = 16;
    } else if (KP_MODEL_TENSOR_DATA_LAYOUT_16W1C8B == model_input_data_layout) {
        width_align_base = 16;
        channel_align_base = 1;
    } else {
        printf("invalid input NPU layout format %s", helper_kp_model_tensor_data_layout_to_string(model_input_data_layout));
        exit(0);
    }

    // calculate width alignment size, channel block count
    uint32_t model_input_width_align = width_align_base * ceil(model_input_width / (float)width_align_base);
    uint32_t model_input_channel_block_num = ceil(model_input_channel / (float)channel_align_base);

    // create re-layout data container
    // KL720 NPU hardware dimension order: CxHxW
    re_layout_buf_size = model_input_channel_block_num * model_input_height *  model_input_width_align * channel_align_base * sizeof(int8_t);
    re_layout_buf = (int8_t *)malloc(re_layout_buf_size);

    // fill data in re-layout data container
    // (for more information, please refer: https://doc.kneron.com/docs/#plus_c/appendix/supported_npu_data_layout_format/)
    uint32_t dst_offset = 0;
    uint32_t src_offset = 0;
    uint32_t model_input_channel_block_idx = 0;
    for (int h = 0; h < model_input_height; h++) {
        for (int w = 0; w < model_input_width; w++) {
            for (int c = 0; c < model_input_channel; c++) {
                model_input_channel_block_idx = c / channel_align_base;

                dst_offset = model_input_channel_block_idx * (model_input_height *  model_input_width_align * channel_align_base) +
                             h * (model_input_width_align * channel_align_base) +
                             w * (channel_align_base) +
                             c % channel_align_base;
                src_offset = h * (img_width * img_channel) +
                             w * (img_channel) +
                             c;

                re_layout_buf[dst_offset] = (int8_t)pre_processing_buf[src_offset];
            }
        }
    }

    free(pre_processing_buf);

    printf("read image ... OK\n");

    printf("\nstarting inference loop %d times:\n", _loop);

    /******* prepare input and output header/buffers *******/
    demo_customize_inf_single_model_header_t input_header;
    demo_customize_inf_single_model_result_t output_result;

    input_header.header_stamp.job_id = DEMO_KL720_CUSTOMIZE_INF_SINGLE_MODEL_JOB_ID;
    input_header.header_stamp.total_image = 1;
    input_header.header_stamp.image_index = 0;
    input_header.width = img_width;
    input_header.height = img_height;

    int header_size = sizeof(demo_customize_inf_single_model_header_t);
    int image_size = img_width * img_height * 4; // RGB565
    int result_size = sizeof(demo_customize_inf_single_model_result_t);
    int recv_size = 0;
//    printf("buffer size %d, width %d, height %d\n",re_layout_buf_size,img_width,img_height);
    /******* starting inference work *******/

    for (int i = 0; i < _loop; i++)
    {
        ret = kp_customized_inference_send(device, (void *)&input_header, header_size, (uint8_t *)re_layout_buf, image_size);
        if (ret != KP_SUCCESS)
            break;

        ret = kp_customized_inference_receive(device, (void *)&output_result, result_size, &recv_size);
        if (ret != KP_SUCCESS)
            break;

        printf("\n[loop %d]\n", i + 1);
        helper_print_yolo_box_on_bmp((kp_yolo_result_t *)&output_result.yolo_result, _image_file_path);
    }
    printf("\n");

    if (ret != KP_SUCCESS)
    {
        printf("\ninference failed, error = %d (%s)\n", ret, kp_error_string(ret));
        return -1;
    }

    free(img_buf);
    kp_disconnect_devices(device);

    return 0;
}
